subroutine jacobi( N, T, dT, maxits, printfrequency, tolerance, &
                   iterations, finalerror)
!-----------------------------------------------------------------------------------
! Apply the Jacobi iteration to a 2D array corresponding to a pure heat diffusion
! problem. The mesh is N X N. If printfrequency > 0, the error norm is printed
! out to a file named "history" every printfrequency iterations. Since the Jacobi
! method is miserably slow, it's best to avoid printing out on every iteration.
!
! The loop order is reversed from the equivalent in C since Fortran stores 2D arrays
! in column-major order. This is not a good solver, but is intended as a starting
! point for improvements via algorithm improvements, multi-threading, and
! parallelism. In particular, the do loops are better replaced by array ops like
!   T = T + dT
! or do concurrent loops. Computing 
!   errnorm = sqrt(sum(abs(dT))/Nsqr)
! is similarly better.
!
! I use the associate() construct to make the loops a little more readable. Also, I
! name the loops (iteration, rows, columns) but that is overkill. It does help when
! doing a red-black update method, but in any case is just a clarity feature.
!
!-----------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!----------------------
! Last Modified: Thu 25 Jan 2018, 06:24 PM
!-----------------------------------------------------------------------------------
    
    use iso_fortran_env, only: stderr => error_unit, real8 => real64
    implicit none
    integer            :: N, maxits, iter, iterations, printfrequency
    integer            :: i, j, history, errorflag
    real(kind=real8)   :: Nsqr, tolerance, errnorm, finalerror, tempsum
    real(kind=real8), target   :: T(0:N+1,0:N+1), dT(0:N+1,0:N+1)
        
    Nsqr = real(N*N, kind=real8)
    errnorm = huge(errnorm)  ! Initialize to largest representable number of the
                             ! same type as errnorm
    iter = 0

    ! Open the output unit for convergence history
    if (printfrequency > 0) then
        open (newunit=history, file='history', action='write', iostat=errorflag)
        if (errorflag /= 0) then
            write(stderr,666) 'Unable to open output file "history"; bailing out'
            stop
        end if
    end if

    associate(                  &
        left  => T(0:N-1,1:N),  & ! Shift first index by  -1  to get point to the left
        right => T(2:N+1,1:N),  & ! Shift first index by  +1  to get point to the right
        up    => T(1:N,2:N+1),  & ! Shift second index by +1  to get point upwards
        down  => T(1:N,0:N-1)   & ! Shift second index by -1  to get point downwards
    )
    iteration: do while (errnorm > tolerance  .and.  iter < maxits)
        tempsum = 0.0_real8
        rows: do j = 1, N
            columns: do i = 1, N
                dT(i,j) = (left(i,j) + right(i,j) + down(i,j) + up(i,j))*0.25_real8  &    
                        - T(i,j)
                tempsum = tempsum + abs(dT(i,j))
            end do columns
        end do rows
    
        do j = 1, N
            do i = 1, N
                T(i,j) = T(i,j) + dT(i,j)
            end do
        end do
        errnorm = tempsum/Nsqr
        iter = iter + 1                
        if (mod(iter, printfrequency) == 1 .and. printfrequency > 0) then 
            write(history,777) iter, errnorm
            flush history
        end if
 
    end do iteration
    end associate
    
    ! Make sure that the last iteration value is printed out, but only once
    if (printfrequency > 0) then 
        if (mod(iter, printfrequency) /= 1) then 
            write(history,777) iter, errnorm
        end if
        close(history)
    end if

    finalerror = errnorm
    iterations = iter

    return
    666 format(a)
    777 format(i0, 4x, es24.17)

end subroutine jacobi
